#define MAX(a, b) (((a) > (b)) ?  (a) : (b))
#define MIN(a, b) (((a) < (b)) ?  (a) : (b))

inline bool _XL_IsFull(PCXLARGE pxl)
{
	if(pxl->end + 1 == pxl->begin) return true;
	if(pxl->end + 1 == pxl->begin + pxl->size) return true;
	return false;
}

inline XL_UINT32 _XL_GetLength(PCXLARGE pxl)
{
	if(pxl->begin < pxl->end) return pxl->end - pxl->begin;
	return (pxl->size - pxl->begin) + pxl->end;
}

inline XL_UINT32* _XL_GetFrac(PCXLARGE pxl, XL_UINT32 i)
{
	if(pxl->begin < pxl->end)
	{
		if(pxl->begin + i < pxl->end)
			return &pxl->frac[pxl->begin + i];
		else
			return nullptr;
	}
	else
	{
		if(pxl->begin + i < pxl->size)
			return &pxl->frac[pxl->begin + i];
		else if(pxl->begin + i - pxl->size < pxl->end)
			return &pxl->frac[pxl->begin + i - pxl->size];
		else
			return nullptr;
	}
}

inline XL_UINT32* _XL_GetFracBack(PCXLARGE pxl)
{
	if(pxl->end) return &pxl->frac[pxl->end - 1];
	else return &pxl->frac[pxl->size - 1];
}

void _XL_Grow(PXLARGE pxl)
{
	if(!pxl) return;
#ifdef XL_DEBUGINFO
	++xl_getdebuginfo()->growCount;
#endif
	XL_UINT32 newSize = (XL_UINT32)(pxl->size * 2);
	if(pxl->begin <= pxl->end)
	{
		pxl->frac = (XL_UINT32*)realloc(pxl->frac, sizeof(XL_UINT32) * newSize);
		pxl->size = newSize;
	}
	else
	{
		XL_UINT32* newFrac = (XL_UINT32*)malloc(sizeof(XL_UINT32) * newSize);
		memcpy(newFrac, &pxl->frac[pxl->begin], sizeof(XL_UINT32) * (pxl->size - pxl->begin));
		memcpy(&newFrac[pxl->size - pxl->begin], pxl->frac, sizeof(XL_UINT32) * pxl->end);
		pxl->end = (pxl->size - pxl->begin) + pxl->end;
		pxl->begin = 0;
		pxl->size = newSize;
		free(pxl->frac);
		pxl->frac = newFrac;
	}
}

void _XL_PushFront(PXLARGE pxl)
{
	if(!pxl) return;
	if(_XL_IsFull(pxl))
	{
		_XL_Grow(pxl);
	}
	
	if(pxl->begin == 0)
	{
		pxl->begin = pxl->size - 1;
		pxl->frac[pxl->begin] = 0;
	}
	else
	{
		pxl->frac[--pxl->begin] = 0;
	}
	--pxl->exp;
}

void _XL_PushFrontN(PXLARGE pxl, XL_UINT32 n)
{
	if(n == 1) return _XL_PushFront(pxl);
	if(!pxl) return;
	while(_XL_GetLength(pxl) + n + 1 > pxl->size)
	{
		_XL_Grow(pxl);
	}
	
	if(pxl->begin < n)
	{
		memset(pxl->frac, 0, sizeof(XL_UINT32) * pxl->begin);
		memset(&pxl->frac[pxl->size - n + pxl->begin], 0, sizeof(XL_UINT32) * (n - pxl->begin));
		pxl->begin = pxl->size - n + pxl->begin;
	}
	else
	{
		memset(&pxl->frac[pxl->begin - n], 0, sizeof(XL_UINT32) * n);
		pxl->begin -= n;
	}
	pxl->exp -= n;
}

int _XL_PopFrontN(PXLARGE pxl, XL_UINT32 n)
{
	if(!pxl) return -1;
	if(_XL_GetLength(pxl) <= n) return -1;
	
	pxl->begin += n;
	if(pxl->begin > pxl->size) pxl->begin -= pxl->size;
	pxl->exp += n;
	return 0;
}

void _XL_PushBack(PXLARGE pxl)
{
	if(!pxl) return;
	if(_XL_IsFull(pxl))
	{
		_XL_Grow(pxl);
	}

	if(pxl->end + 1 == pxl->size)
	{
		pxl->end = 0;
		pxl->frac[pxl->size - 1] = 0;
	}
	else
	{
		pxl->frac[pxl->end++] = 0;
	}
}

int _XL_PopBack(PXLARGE pxl)
{
	if(!pxl) return -1;
	if(_XL_GetLength(pxl) <= 1) return -1;
	
	if(pxl->end == 0)
	{
		pxl->end = pxl->size - 1;
	}
	else
	{
		--pxl->end;
	}
	return 0;
}

void _XL_PushBackN(PXLARGE pxl, XL_UINT32 n)
{
	if(n == 1) return _XL_PushBack(pxl);
	if(!pxl) return;
	while(_XL_GetLength(pxl) + n + 1 > pxl->size)
	{
		_XL_Grow(pxl);
	}

	if(pxl->end + n >= pxl->size)
	{
		memset(&pxl->frac[pxl->end], 0, sizeof(XL_UINT32) * (pxl->size - pxl->end));
		memset(pxl->frac, 0, sizeof(XL_UINT32) * (n - pxl->size + pxl->end));
		pxl->end = n - pxl->size + pxl->end;
	}
	else
	{
		memset(&pxl->frac[pxl->end], 0, sizeof(XL_UINT32) * n);
		pxl->end += n;
	}
}

void _XL_Shrink(PXLARGE pxl)
{
	if(!pxl) return;
	while(pxl->frac[(pxl->end == 0) ? pxl->size - 1 : pxl->end - 1] == 0)
	{
		if(pxl->begin + 1 == pxl->end) break;
		if(pxl->begin + 1 == pxl->end + pxl->size) break;
		if(pxl->end == 0) pxl->end = pxl->size - 1;
		else --pxl->end;
	}

	while(pxl->frac[pxl->begin] == 0)
	{
		if(pxl->begin + 1 == pxl->end) break;
		if(pxl->begin + 1 == pxl->end + pxl->size) break;
		++pxl->begin;
		if(pxl->begin == pxl->size) pxl->begin = 0;
		++pxl->exp;
	}

	if(pxl->frac[pxl->begin] == 0)
	{
		pxl->exp = 0;
	}
}

inline void _XL_ExpandCopy(PXLARGE dest, PCXLARGE src, XL_UINT32 lowerSize, XL_UINT32 upperSize)
{
	XL_UINT32 length = _XL_GetLength(src);
	XL_UINT32 eLength = length + lowerSize + upperSize;
	// ũⰡ   
	if (eLength + 1 >= dest->size)
	{
		free(dest->frac);
		dest->frac = (XL_UINT32*)malloc(sizeof(XL_UINT32)* (eLength + 1));
		dest->size = eLength + 1;
	}

	if (lowerSize > 0)
	{
		memset(dest->frac, 0, sizeof(XL_UINT32)*lowerSize);
	}

	if (src->begin <= src->end)
	{
		memcpy(dest->frac + lowerSize, &src->frac[src->begin], sizeof(XL_UINT32)* length);
	}
	else
	{
		memcpy(dest->frac + lowerSize, &src->frac[src->begin], sizeof(XL_UINT32)* (src->size - src->begin));
		memcpy(&dest->frac[lowerSize + src->size - src->begin], src->frac, sizeof(XL_UINT32)* src->end);
	}

	if (upperSize > 0)
	{
		memset(dest->frac + lowerSize + length, 0, sizeof(XL_UINT32)*upperSize);
	}

	dest->begin = 0;
	dest->end = eLength;
	dest->exp = src->exp - lowerSize;
	dest->sign = src->sign;
}

inline void _XL_GetData(XL_UINT32* dest, PXLARGE src, XL_UINT32 begin, XL_UINT32 end)
{
	memcpy(dest, src->frac + src->begin, sizeof(XL_UINT32)*(end - begin));
}

inline void _XL_GetDataW(XL_UINT32* dest, PXLARGE src, XL_INT32 beginExp, XL_INT32 endExp)
{
	XL_UINT32 sLen = _XL_GetLength(src);
	XL_INT32 srcEnd = src->exp + sLen;

	if (endExp < srcEnd)
	{
		if (endExp < src->exp)
		{
			memset(dest, 0, sizeof(XL_UINT32)*(endExp - beginExp));
		}
		else if (beginExp < src->exp)
		{
			memset(dest, 0, sizeof(XL_UINT32)*(src->exp - beginExp));
			_XL_GetData(dest + (src->exp - beginExp), src, 0, endExp - src->exp);
		}
		else
		{
			_XL_GetData(dest, src, beginExp - src->exp, endExp - src->exp);
		}
	}
	else if (beginExp < srcEnd)
	{
		if (beginExp < src->exp)
		{
			memset(dest, 0, sizeof(XL_UINT32)*(src->exp - beginExp));
			_XL_GetData(dest + (src->exp - beginExp), src, 0, sLen);
			memset(dest + (srcEnd - beginExp), 0, sizeof(XL_UINT32)*(endExp - srcEnd));
		}
		else
		{
			_XL_GetData(dest, src, 0, srcEnd - beginExp);
			memset(dest + (srcEnd - beginExp), 0, sizeof(XL_UINT32)*(endExp - srcEnd));
		}
	}
	else
	{
		memset(dest, 0, sizeof(XL_UINT32)*(endExp - beginExp));
	}
}

inline void _XL_FracRestLower(PXLARGE dest, PCXLARGE src, XL_UINT32 n)
{
	dest->frac = src->frac;
	dest->size = src->size;
	dest->exp = 0;
	dest->sign = 0;
	dest->begin = (src->begin + n) % src->size;
	dest->end = src->end;
}

inline void _XL_FracLower(PXLARGE dest, PCXLARGE src, XL_UINT32 n)
{
	dest->frac = src->frac;
	dest->size = src->size;
	dest->exp = 0;
	dest->sign = 0;
	dest->begin = src->begin;
	dest->end = (src->begin + n) % src->size;
}

// _XL_DivideFrac : divide src into two parts, [begin, pos) [pos, end)
inline void _XL_DivideFrac(PXLARGE lower, PXLARGE upper, PCXLARGE src, XL_UINT32 pos)
{
	lower->frac = upper->frac = src->frac;
	lower->size = upper->size = src->size;
	lower->exp = upper->exp = 0;
	lower->sign = upper->sign = 0;
	lower->begin = src->begin;
	lower->end = upper->begin = (src->begin + pos) % src->size;
	upper->end = src->end;
}

inline void _XL_FracRestUpper(PXLARGE dest, PCXLARGE src, XL_UINT32 n)
{
	dest->frac = src->frac;
	dest->size = src->size;
	dest->exp = 0;
	dest->sign = 0;
	dest->begin = src->begin;
	dest->end = (src->end + src->size - n) % src->size;
}

inline void _XL_FracUpper(PXLARGE dest, PCXLARGE src, XL_UINT32 n)
{
	dest->frac = src->frac;
	dest->size = src->size;
	dest->exp = 0;
	dest->sign = 0;
	dest->begin = (src->end + src->size - n) % src->size;
	dest->end = src->end;
}