#include "stdafx.h"
#include "XLarge.h"
#include <intrin.h>
#include "XLCalcUtils.h"
#include "XLQueueUtils.h"

#define XL_TIMER_ON

#include "CPUTimer.h"

#pragma intrinsic(__emulu)

#define MAX(a, b) (((a) > (b)) ?  (a) : (b))
#define MIN(a, b) (((a) < (b)) ?  (a) : (b))

inline void DEBUG_Print(const char* before, PCXLARGE pxl, const char* after)
{
#ifdef _DEBUG
	printf(before);
	xl_print(pxl, stdout, -1);
	printf(after);
#endif
}

#ifdef XL_DEBUGINFO
XLDebugInfo* xl_getdebuginfo()
{
	static XLDebugInfo di;
	return &di;
}

#endif

#define K_THRESHOLD 48

PXLARGE xl_create(XL_UINT32 size)
{
	if(size <= 1) return nullptr;
	PXLARGE pxl = (PXLARGE)malloc(sizeof(XLarge));
	pxl->size = size;
	pxl->frac = (XL_UINT32*)malloc(sizeof(XL_UINT32) * size);
	pxl->begin = 0;
	pxl->end = 1;
	pxl->exp = 0;
	pxl->sign = 0;
	return pxl;
}

PXLARGE xl_createcopy(PCXLARGE src)
{
	PXLARGE pxl = (PXLARGE)malloc(sizeof(XLarge));
	pxl->size = src->size;
	pxl->frac = (XL_UINT32*)malloc(sizeof(XL_UINT32) * src->size);
	xl_move(pxl, src);
	return pxl;
}

void xl_delete(PXLARGE pxl)
{
	if(!pxl) return;
	free(pxl->frac);
	free(pxl);
}

void xl_move(PXLARGE dest, PCXLARGE src)
{
	_XL_ExpandCopy(dest, src, 0, 0);
}

void xl_init(PXLARGE dest, XL_INT32 value)
{
	dest->frac[0] = abs(value);
	dest->begin = 0;
	dest->end = 1;
	dest->exp = 0;
	dest->sign = value < 0;
}

void xl_init1(PXLARGE dest, XL_UINT32 value)
{
	dest->frac[0] = value;
	dest->begin = 0;
	dest->end = 1;
	dest->exp = 0;
	dest->sign = 0;
}

void xl_swap(PXLARGE a, PXLARGE b)
{
	XLarge t = *a;
	*a = *b;
	*b = t;
}

void xl_setnan(PXLARGE pxl)
{
	pxl->sign = (XL_UINT32)-1;
}

int xl_isnan(PCXLARGE pxl)
{
	return pxl->sign == (XL_UINT32)-1;
}

inline void _Abs_Add(PXLARGE dest, PCXLARGE src)
{
	XL_UINT32 srcLen = _XL_GetLength(src);
	XL_UINT32 expBias = src->exp - dest->exp;
	if(dest->exp > src->exp)
	{
		expBias = 0;
		// 아랫쪽으로 확장
		_XL_PushFrontN(dest, dest->exp - src->exp);
	}
	XL_INT32 backIncrease = (src->exp + srcLen) - (dest->exp + _XL_GetLength(dest));
	if(backIncrease > 0) _XL_PushBackN(dest, backIncrease + 1);
	else _XL_PushBack(dest);

	XL_UINT32 orgDestBegin = dest->begin;
	//XL_UINT32 orgDestLen = _XL_GetLength(dest);
	
	// 자릿수를 맞추기 위해 시작지점 변경
	dest->begin += expBias;
	if(dest->begin >= dest->size) dest->begin -= dest->size;
	XL_UINT32 destLen = _XL_GetLength(dest);

	if(dest->begin < dest->end || dest->begin + srcLen < dest->size)
	{
		int carry = 0;
		if(src->begin < src->end)
		{
			carry = i_adds(&dest->frac[dest->begin], &src->frac[src->begin], srcLen, 0);
		}
		else
		{
			carry = i_adds(&dest->frac[dest->begin], &src->frac[src->begin], src->size - src->begin, 0);
			carry = i_adds(&dest->frac[dest->begin + src->size - src->begin], src->frac, src->end, carry);
		}
		// 위쪽으로 받아올림 있을 경우 받아올림 처리
		if(carry)
		{
			if(dest->begin < dest->end)
			{
				i_addu32(&dest->frac[dest->begin + srcLen], destLen - srcLen, 1);
			}
			else
			{
				if(i_addu32(&dest->frac[dest->begin + srcLen], dest->size - dest->begin - srcLen, 1))
				{
					if(dest->end) i_addu32(dest->frac, dest->end, 1);
				}
			}
		}
	}
	else
	{
		int carry = 0;
		XL_UINT32 destSep = dest->size - dest->begin;
		if(src->begin < src->end)
		{
			carry = i_adds(&dest->frac[dest->begin], &src->frac[src->begin], destSep, 0);
			carry = i_adds(dest->frac, &src->frac[src->begin + destSep], srcLen - destSep, carry);
		}
		else
		{
			XL_UINT32 srcSep = src->size - src->begin;
			if(destSep == srcSep)
			{
				carry = i_adds(&dest->frac[dest->begin], &src->frac[src->begin], destSep, 0);
				carry = i_adds(dest->frac, src->frac, srcLen - destSep, carry);
			}
			else if(destSep < srcSep)
			{
				carry = i_adds(&dest->frac[dest->begin], &src->frac[src->begin], destSep, 0);
				carry = i_adds(dest->frac, &src->frac[src->begin + destSep], srcSep - destSep, carry);
				carry = i_adds(&dest->frac[srcSep - destSep], src->frac, srcLen - srcSep, carry);
			}
			else
			{
				carry = i_adds(&dest->frac[dest->begin], &src->frac[src->begin], srcSep, 0);
				carry = i_adds(&dest->frac[dest->begin + srcSep], src->frac, destSep - srcSep, carry);
				carry = i_adds(dest->frac, &src->frac[destSep - srcSep], srcLen - destSep, carry);
			}
		}
		// 위쪽으로 받아올림 있을 경우 받아올림 처리
		if(carry)
		{
			i_addu32(&dest->frac[srcLen - destSep], dest->end - srcLen + destSep, 1);
		}
	}
	// 변경한 시작지점 복원
	dest->begin = orgDestBegin;
	_XL_Shrink(dest);
}

inline void _Abs_Sub(PXLARGE dest, PCXLARGE src)
{
	XL_UINT32 srcLen = _XL_GetLength(src);
	XL_UINT32 expBias = src->exp - dest->exp;
	if(dest->exp > src->exp)
	{
		expBias = 0;
		// 아랫쪽으로 확장
		_XL_PushFrontN(dest, dest->exp - src->exp);
	}
	XL_INT32 backIncrease = (src->exp + srcLen) - (dest->exp + _XL_GetLength(dest));
	if(backIncrease > 0) _XL_PushBackN(dest, backIncrease + 1);
	else _XL_PushBack(dest);

	XL_UINT32 orgDestBegin = dest->begin;
	XL_UINT32 orgDestLen = _XL_GetLength(dest);
	
	// 자릿수를 맞추기 위해 시작지점 변경
	dest->begin += expBias;
	if(dest->begin >= dest->size) dest->begin -= dest->size;
	XL_UINT32 destLen = _XL_GetLength(dest);
	
	bool invert = false;

	if(dest->begin < dest->end || dest->begin + srcLen < dest->size)
	{
		int carry = 0;
		if(src->begin < src->end)
		{
			carry = i_subs(&dest->frac[dest->begin], &src->frac[src->begin], srcLen, 0);
		}
		else
		{
			carry = i_subs(&dest->frac[dest->begin], &src->frac[src->begin], src->size - src->begin, 0);
			carry = i_subs(&dest->frac[dest->begin + src->size - src->begin], src->frac, src->end, carry);
		}

		// 위쪽으로 받아내림 있을 경우 받아내림 처리
		if(carry)
		{
			if(destLen == srcLen)
			{
				invert = true;
			}
			else if(dest->begin < dest->end)
			{
				if(i_subu32(&dest->frac[dest->begin + srcLen], destLen - srcLen, 1))
				{
					invert = true;
				}
			}
			else
			{
				if(i_subu32(&dest->frac[dest->begin + srcLen], dest->size - dest->begin - srcLen, 1))
				{
					if(i_subu32(dest->frac, dest->end, 1))
					{
						invert = true;
					}
				}
			}
		}
	}
	else
	{
		int carry = 0;
		XL_UINT32 destSep = dest->size - dest->begin;
		if(src->begin < src->end)
		{
			carry = i_subs(&dest->frac[dest->begin], &src->frac[src->begin], destSep, 0);
			carry = i_subs(dest->frac, &src->frac[src->begin + destSep], srcLen - destSep, carry);
		}
		else
		{
			XL_UINT32 srcSep = src->size - src->begin;
			if(destSep == srcSep)
			{
				carry = i_subs(&dest->frac[dest->begin], &src->frac[src->begin], destSep, 0);
				carry = i_subs(dest->frac, src->frac, srcLen - destSep, carry);
			}
			else if(destSep < srcSep)
			{
				carry = i_subs(&dest->frac[dest->begin], &src->frac[src->begin], destSep, 0);
				carry = i_subs(dest->frac, &src->frac[src->begin + destSep], srcSep - destSep, carry);
				carry = i_subs(&dest->frac[srcSep - destSep], src->frac, srcLen - srcSep, carry);
			}
			else
			{
				carry = i_subs(&dest->frac[dest->begin], &src->frac[src->begin], srcSep, 0);
				carry = i_subs(&dest->frac[dest->begin + srcSep], src->frac, destSep - srcSep, carry);
				carry = i_subs(dest->frac, &src->frac[destSep - srcSep], srcLen - destSep, carry);
				
			}
		}
		// 위쪽으로 받아내림 있을 경우 받아내림 처리
		if(carry)
		{
			if(i_subu32(&dest->frac[srcLen - destSep], dest->end - srcLen + destSep, 1))
			{
				invert = true;
			}
		}
	}
	// 변경한 시작지점 복원
	dest->begin = orgDestBegin;

	// 부호 뒤집기
	if(invert)
	{
		dest->sign = !dest->sign;
		if(dest->begin < dest->end)
		{
			i_cmplt(&dest->frac[dest->begin], orgDestLen, 0);
		}
		else
		{
			int carry = i_cmplt(&dest->frac[dest->begin], dest->size - dest->begin, 0);
			if(dest->end) i_cmplt(dest->frac, dest->end, carry);
		}
	}
	_XL_Shrink(dest);
}

void xl_add(PXLARGE dest, PCXLARGE src)
{
	if(xl_isnan(dest) || xl_isnan(src)) return xl_setnan(dest);
	return (dest->sign == src->sign ? _Abs_Add : _Abs_Sub)(dest, src);
}

void xl_sub(PXLARGE dest, PCXLARGE src)
{
	if(xl_isnan(dest) || xl_isnan(src)) return xl_setnan(dest);
	return (dest->sign != src->sign ? _Abs_Add : _Abs_Sub)(dest, src);
}

inline void _Abs_Add1(PXLARGE dest, XL_UINT32 src)
{
	if (dest->exp > 0)
	{
		_XL_PushFrontN(dest, dest->exp);
	}
	XL_INT32 backIncrease = 1 - (dest->exp + _XL_GetLength(dest));
	if (backIncrease > 0) _XL_PushBackN(dest, backIncrease + 1);
	else _XL_PushBack(dest);

	if (dest->begin < dest->end)
	{
		i_addu32(&dest->frac[dest->begin], dest->end - dest->begin, src);
	}
	else
	{
		int carry = 0;
		XL_UINT32 destSep = dest->size - dest->begin;
		carry = i_addu32(&dest->frac[dest->begin], destSep, src);
		i_addu32(dest->frac, dest->end , carry);
	}
	_XL_Shrink(dest);
}

inline void _Abs_Sub1(PXLARGE dest, XL_UINT32 src)
{
	if (dest->exp > 0)
	{
		_XL_PushFrontN(dest, dest->exp);
	}
	XL_INT32 backIncrease = 1 - (dest->exp + _XL_GetLength(dest));
	if (backIncrease > 0) _XL_PushBackN(dest, backIncrease);

	bool invert = false;

	if (dest->begin < dest->end)
	{
		invert = !!i_subu32(&dest->frac[dest->begin], dest->end - dest->begin, src);
	}
	else
	{
		int carry = 0;
		XL_UINT32 destSep = dest->size - dest->begin;
		carry = i_subu32(&dest->frac[dest->begin], destSep, src);
		invert = !!i_subu32(dest->frac, dest->end, carry);
	}

	// 부호 뒤집기
	if (invert)
	{
		dest->sign = !dest->sign;
		if (dest->begin < dest->end)
		{
			i_cmplt(&dest->frac[dest->begin], dest->end - dest->begin, 0);
		}
		else
		{
			int carry = i_cmplt(&dest->frac[dest->begin], dest->size - dest->begin, 0);
			if (dest->end) i_cmplt(dest->frac, dest->end, carry);
		}
	}

	_XL_Shrink(dest);
}

void xl_add1(PXLARGE dest, XL_UINT32 src)
{
	return (dest->sign == 0 ? _Abs_Add1 : _Abs_Sub1)(dest, src);
}

void xl_sub1(PXLARGE dest, XL_UINT32 src)
{
	return (dest->sign == 0 ? _Abs_Sub1 : _Abs_Add1)(dest, src);
}

void xl_neg(PXLARGE dest)
{
	if(xl_isnan(dest)) return;
	dest->sign = !dest->sign;
}

void xl_printx(PCXLARGE pxl, FILE* file)
{
	XL_UINT32 size = _XL_GetLength(pxl);
	if(pxl->sign)
	{
		fprintf(file, "-");
	}
	if(pxl->exp >= 0)
	{
		XL_UINT32 i = size - 1;
		fprintf(file, "%X", *_XL_GetFrac(pxl, i--));
		for(;i != (XL_UINT32)-1; --i)
		{
			fprintf(file, "%08X", *_XL_GetFrac(pxl, i));
		}
		for(int j = 0; j < pxl->exp; ++j)
		{
			fprintf(file, "00000000");
		}
	}
	else
	{

	}
}

void _XL_MulN(PXLARGE dest, PCXLARGE a, PCXLARGE b)
{
	XL_UINT32 aLen = _XL_GetLength(a);
	XL_UINT32 bLen = _XL_GetLength(b);
	xl_init(dest, 0);
	_XL_PushBackN(dest, aLen + bLen - 1);
	
	for(XL_UINT32 i = 0; i < aLen; ++i) for(XL_UINT32 j = 0; j < bLen; ++j)
	{
		XL_UINT32 ad = *_XL_GetFrac(a, i);
		XL_UINT32 bd = *_XL_GetFrac(b, j);
		if(ad && bd)
		{
			i_addu64(&dest->frac[dest->begin + i + j], dest->end - dest->begin - i - j, __emulu(ad, bd));
		}
	}
	_XL_Shrink(dest);
	dest->exp += a->exp + b->exp;
}

void _XL_MulK(PXLARGE dest, PCXLARGE a, PCXLARGE b)
{
	XL_UINT32 aLen = _XL_GetLength(a);
	XL_UINT32 bLen = _XL_GetLength(b);
	if(MIN(aLen, bLen) < K_THRESHOLD)
	{
		return _XL_MulN(dest, a, b);
	}

	if((aLen + 1) / 2 + 4 >= bLen)
	{
		XL_UINT32 half = (aLen + 1) / 2;
		XLarge a0, a1;
		_XL_FracLower(&a0, a, half);
		_XL_FracRestLower(&a1, a, half);
		
		PXLARGE z1 = xl_create(aLen + 1);
		_XL_MulK(dest, &a0, b);
		_XL_MulK(z1, &a1, b);
		z1->exp += half;
		xl_add(dest, z1);
		xl_delete(z1);
		
		// 사라진 a의 지수부 복원
		dest->exp += a->exp;
	}
	else if(aLen <= (bLen + 1) / 2 + 4)
	{
		XL_UINT32 half = (bLen + 1) / 2;
		XLarge b0, b1;
		_XL_FracLower(&b0, b, half);
		_XL_FracRestLower(&b1, b, half);
		
		PXLARGE z1 = xl_create(bLen + 1);
		_XL_MulK(dest, a, &b0);
		_XL_MulK(z1, a, &b1);
		z1->exp += half;
		xl_add(dest, z1);
		xl_delete(z1);
		
		// 사라진 b의 지수부 복원
		dest->exp += b->exp;
	}
	else
	{
		XL_UINT32 half = (MAX(aLen, bLen) + 1) / 2;
		XLarge a0, a1, b0, b1;
		_XL_FracLower(&a0, a, half);
		_XL_FracLower(&b0, b, half);
		_XL_FracRestLower(&a1, a, half);
		_XL_FracRestLower(&b1, b, half);

		PXLARGE z0 = xl_create(2*half);
		PXLARGE z1 = xl_create(2*half);
		PXLARGE z2 = xl_create(2*half);

		xl_move(z0, &a0);
		xl_add(z0, &a1);
		xl_move(z2, &b0);
		xl_add(z2, &b1);

		_XL_MulK(z1, z0, z2); // z1 = (a0 + a1) * (b0 + b1)
		_XL_MulK(z0, &a0, &b0); // z0 = a0 * b0
		_XL_MulK(z2, &a1, &b1); // z2 = a1 * b1
		xl_sub(z1, z0);
		xl_sub(z1, z2);

		z2->exp += 2 * half;
		z1->exp += half;

		xl_move(dest, z0);
		xl_add(dest, z1);
		xl_add(dest, z2);

		xl_delete(z2);
		xl_delete(z1);
		xl_delete(z0);

		// 사라진 a, b의 지수부 복원
		dest->exp += a->exp + b->exp;
	}
	_XL_Shrink(dest);
}

void xl_mul(PXLARGE dest, PCXLARGE a, PCXLARGE b)
{
	if(xl_isnan(a) || xl_isnan(b)) return xl_setnan(dest);
	_XL_MulK(dest, a, b);
	dest->sign = a->sign ^ b->sign;
}

void xl_mul1(PXLARGE dest, PCXLARGE a, XL_UINT32 b)
{
	if (xl_isnan(a)) return xl_setnan(dest);
	XL_UINT32 aLen = _XL_GetLength(a);
	xl_init(dest, 0);
	_XL_PushBackN(dest, aLen);

	for (XL_UINT32 i = 0; i < aLen; ++i)
	{
		XL_UINT32 ad = *_XL_GetFrac(a, i);
		if (ad)
		{
			i_addu64(&dest->frac[dest->begin + i], dest->end - dest->begin - i, __emulu(ad, b));
		}
	}
	_XL_Shrink(dest);
	dest->exp += a->exp;
}

inline void _Abs_MulAdd(PXLARGE dest, PCXLARGE a, PCXLARGE b)
{
	XL_UINT32 aLen = _XL_GetLength(a);
	XL_UINT32 bLen = _XL_GetLength(b);
	XL_INT32 tExp = a->exp + b->exp;
	XL_INT32 tBack = tExp + aLen + bLen + 1; // carry를 대비한 여분의 +1
	if(dest->exp > tExp) _XL_PushFrontN(dest, dest->exp - tExp);
	XL_INT32 dBack = dest->exp + _XL_GetLength(dest);
	if(dBack < tBack) _XL_PushBackN(dest, tBack - dBack);

	XL_UINT32 orgDestBegin = dest->begin;
	if(tExp > dest->exp) dest->begin += tExp - dest->exp;
	if(dest->begin > dest->size) dest->begin -= dest->size;

	if(dest->begin < dest->end)
	{
		for(XL_UINT32 i = 0; i < aLen; ++i) for(XL_UINT32 j = 0; j < bLen; ++j)
		{
			XL_UINT32 ad = *_XL_GetFrac(a, i);
			XL_UINT32 bd = *_XL_GetFrac(b, j);
			if(ad && bd)
			{
				i_addu64(&dest->frac[dest->begin + i + j], dest->end - dest->begin - i - j, __emulu(ad, bd));
			}
		}
	}
	else
	{
		for(XL_UINT32 i = 0; i < aLen; ++i) for(XL_UINT32 j = 0; j < bLen; ++j)
		{
			XL_UINT32 ad = *_XL_GetFrac(a, i);
			XL_UINT32 bd = *_XL_GetFrac(b, j);
			if(ad && bd)
			{
				if(dest->begin + i + j < dest->size)
				{
					i_addu64_2(&dest->frac[dest->begin + i + j], dest->size - dest->begin - i - j, dest->frac, dest->end, __emulu(ad, bd));
				}
				else
				{
					i_addu64(&dest->frac[dest->begin + i + j - dest->size], dest->end + dest->size - dest->begin - i - j, __emulu(ad, bd));
				}
			}
		}
	}
	dest->begin = orgDestBegin;
	_XL_Shrink(dest);
}

inline void _Abs_MulSub(PXLARGE dest, PCXLARGE a, PCXLARGE b)
{
	XL_UINT32 aLen = _XL_GetLength(a);
	XL_UINT32 bLen = _XL_GetLength(b);
	XL_INT32 tExp = a->exp + b->exp;
	XL_INT32 tBack = tExp + aLen + bLen;
	if(dest->exp > tExp) _XL_PushFrontN(dest, dest->exp - tExp);
	XL_INT32 dBack = dest->exp + _XL_GetLength(dest);
	if(dBack < tBack) _XL_PushBackN(dest, tBack - dBack);

	XL_UINT32 orgDestBegin = dest->begin;
	if(tExp > dest->exp) dest->begin += tExp - dest->exp;
	if(dest->begin > dest->size) dest->begin -= dest->size;

	bool invert = false;

	if(dest->begin < dest->end)
	{
		for(XL_UINT32 i = 0; i < aLen; ++i) for(XL_UINT32 j = 0; j < bLen; ++j)
		{
			XL_UINT32 ad = *_XL_GetFrac(a, i);
			XL_UINT32 bd = *_XL_GetFrac(b, j);
			if(ad && bd)
			{
				if((invert ? i_addu64 : i_subu64)(&dest->frac[dest->begin + i + j], dest->end - dest->begin - i - j, __emulu(ad, bd)))
				{
					// 부호 뒤집기
					invert = true;
					dest->sign = !dest->sign;
					if(orgDestBegin < dest->end)
					{
						i_cmplt(&dest->frac[orgDestBegin], dest->end - orgDestBegin, 0);
					}
					else
					{
						int carry = i_cmplt(&dest->frac[orgDestBegin], dest->size - orgDestBegin, 0);
						if(dest->end) i_cmplt(dest->frac, dest->end, carry);
					}
					
				}
			}
		}
	}
	else
	{
		for(XL_UINT32 i = 0; i < aLen; ++i) for(XL_UINT32 j = 0; j < bLen; ++j)
		{
			XL_UINT32 ad = *_XL_GetFrac(a, i);
			XL_UINT32 bd = *_XL_GetFrac(b, j);
			if(ad && bd)
			{
				if(dest->begin + i + j < dest->size)
				{
					if((invert ? i_addu64_2 : i_subu64_2)(&dest->frac[dest->begin + i + j], dest->size - dest->begin - i - j, dest->frac, dest->end, __emulu(ad, bd)))
					{
						// 부호 뒤집기
						invert = true;
						dest->sign = !dest->sign;
						int carry = i_cmplt(&dest->frac[orgDestBegin], dest->size - orgDestBegin, 0);
						if(dest->end) i_cmplt(dest->frac, dest->end, carry);
					}
				}
				else
				{
					if((invert ? i_addu64 : i_subu64)(&dest->frac[dest->begin + i + j - dest->size], dest->end + dest->size - dest->begin - i - j, __emulu(ad, bd)))
					{
						// 부호 뒤집기
						invert = true;
						dest->sign = !dest->sign;
						int carry = i_cmplt(&dest->frac[orgDestBegin], dest->size - orgDestBegin, 0);
						if(dest->end) i_cmplt(dest->frac, dest->end, carry);
					}
				}
			}
		}
	}
	dest->begin = orgDestBegin;
	_XL_Shrink(dest);
}

void xl_muladd(PXLARGE dest, PCXLARGE a, PCXLARGE b)
{
	if(xl_isnan(dest) || xl_isnan(a) || xl_isnan(b)) return xl_setnan(dest);
	if(MIN(_XL_GetLength(a), _XL_GetLength(b)) < K_THRESHOLD)
		return (dest->sign == (a->sign ^ b->sign) ? _Abs_MulAdd : _Abs_MulSub)(dest, a, b);
	PXLARGE t = xl_create(_XL_GetLength(a) + _XL_GetLength(b));
	xl_mul(t, a, b);
	xl_add(dest, t);
	xl_delete(t);
}

void xl_mulsub(PXLARGE dest, PCXLARGE a, PCXLARGE b)
{
	if(xl_isnan(dest) || xl_isnan(a) || xl_isnan(b)) return xl_setnan(dest);
	if(MIN(_XL_GetLength(a), _XL_GetLength(b)) < K_THRESHOLD)
		return (dest->sign != (a->sign ^ b->sign) ? _Abs_MulAdd : _Abs_MulSub)(dest, a, b);
	PXLARGE t = xl_create(_XL_GetLength(a) + _XL_GetLength(b));
	xl_mul(t, a, b);
	xl_sub(dest, t);
	xl_delete(t);
}

inline void _XL_SHL(PXLARGE dest, XL_UINT32 bit)
{
	dest->exp += bit >> 5;
	if (bit & 31)
	{
		_XL_PushBack(dest);
		if (dest->begin < dest->end)
		{
			i_shls(&dest->frac[dest->begin], dest->end - dest->begin, bit & 31, 0);
		}
		else
		{
			XL_UINT32 carry = i_shls(&dest->frac[dest->begin], dest->size - dest->begin, bit & 31, 0);
			if (dest->end) i_shls(dest->frac, dest->end, bit & 31, carry);
		}
	}
}

void xl_shl(PXLARGE dest, XL_UINT32 bit)
{
	if(xl_isnan(dest)) return;
	_XL_SHL(dest, bit);
	_XL_Shrink(dest);
}

void xl_shr(PXLARGE dest, XL_UINT32 bit)
{
	if(xl_isnan(dest)) return;
	dest->exp -= bit >> 5;
	if(bit & 31)
	{
		_XL_PushFront(dest);
		if(dest->begin < dest->end)
		{
			i_shrs(&dest->frac[dest->begin], dest->end - dest->begin, bit & 31, 0);
		}
		else
		{
			XL_UINT32 carry = 0;
			if(dest->end) carry = i_shrs(dest->frac, dest->end, bit & 31, 0);
			i_shrs(&dest->frac[dest->begin], dest->size - dest->begin, bit & 31, carry);
		}
	}
	_XL_Shrink(dest);
}

inline int _Abs_Cmp(PCXLARGE a, PCXLARGE b)
{
	XL_INT32 aLen = _XL_GetLength(a);
	XL_INT32 bLen = _XL_GetLength(b);
	if(a->exp + aLen < b->exp + bLen) return -1;
	if(a->exp + aLen > b->exp + bLen) return 1;
	
	XL_INT32 mExp = MAX(a->exp, b->exp);

	XL_UINT32* a1 = &a->frac[a->begin + mExp - a->exp];
	XL_UINT32* a2 = nullptr;
	XL_UINT32 aSep = MIN(aLen, bLen);
	XL_UINT32* b1 = &b->frac[b->begin + mExp - b->exp];
	XL_UINT32* b2 = nullptr;
	XL_UINT32 bSep = MIN(aLen, bLen);

	if(a1 > a->frac + a->size)
	{
		a1 -= a->size;
	}
	else if(a->begin > a->end)
	{
		a2 = a->frac;
		aSep = a->size - (a->begin + mExp - a->exp);
	}

	if(b1 > b->frac + b->size)
	{
		b1 -= b->size;
	}
	else if(b->begin > b->end)
	{
		b2 = b->frac;
		bSep = b->size - (b->begin + mExp - b->exp);
	}

	int r = i_cmps2(a1, aSep, a2, b1, bSep, b2, MIN(aLen, bLen));
	if(r) return r;
	if(a->exp > b->exp) return -1;
	if(a->exp < b->exp) return 1;
	return 0;
}


int xl_cmp(PCXLARGE a, PCXLARGE b)
{
	if(xl_isnan(a) || xl_isnan(b)) return -1;
	if(a->sign == 0)
	{
		if(b->sign == 0) return _Abs_Cmp(a, b);
		else return 1;
	}
	else
	{
		if(b->sign == 0) return -1;
		else return -_Abs_Cmp(a, b);
	}
}

int xl_cmpzero(PCXLARGE a)
{
	if(xl_isnan(a)) return -1;
	if(*_XL_GetFracBack(a) == 0 && _XL_GetLength(a) == 1) return 0;
	if(a->sign == 0) return 1;
	return -1;
}

int xl_cmpnearzero(PCXLARGE a, XL_INT32 exp)
{
	int r = xl_cmpzero(a);
	if(r == 0) return 0;
	if((XL_INT32)_XL_GetLength(a) + a->exp < exp) return 0;
	return r;
}

void _XL_ModNMN(PCXLARGE a, PCXLARGE b, PXLARGE q, PXLARGE r)
{
	XL_UINT32 bs = i_bsu(*_XL_GetFracBack(b));
	if (bs == (XL_UINT32)(-1))
	{
		// b 가 0인 경우
		xl_setnan(q);
		xl_setnan(r);
		return;
	}

	switch (_Abs_Cmp(a, b))
	{
	case 0:
		xl_init1(q, 1);
		xl_init1(r, 0);
		return;
	case -1:
		xl_init1(q, 0);
		xl_move(r, a);
		return;
	}

	_XL_ExpandCopy(r, a, MAX(a->exp - b->exp, 0), 1);
	PXLARGE nb = xl_createcopy(b);

	if (bs < 31)
	{
		_XL_SHL(r, 31 - bs);
		xl_shl(nb, 31 - bs);
	}

	if (*_XL_GetFracBack(r) == 0)
	{
		_XL_PopBack(r);
	}

	XL_INT32 expDif = nb->exp - r->exp;

	XL_UINT32 alen = _XL_GetLength(r);
	XL_UINT32 blen = _XL_GetLength(nb);
	xl_init1(q, 0);
	_XL_PushBackN(q, alen - blen);
	if (i_cmps(&r->frac[r->end - blen], &nb->frac[nb->begin], blen) >= 0)
	{
		q->frac[q->begin + alen - (blen + expDif)] = 1;
		i_subs(&r->frac[r->end - blen], &nb->frac[nb->begin], blen, 0);
	}

	nb->exp = 0;
	DEBUG_Print("nb: ", nb, "\n");
	PXLARGE t = xl_create(alen);
	for (XL_INT32 i = alen - 2; i >= (XL_INT32)(blen - 1 + expDif); --i)
	{
		XL_UINT32 quot;
		i_div64(*(XL_UINT64*)(r->frac + i), *_XL_GetFracBack(nb), &quot);
		xl_mul1(t, nb, quot);
		DEBUG_Print("t: ", t, "\n");
		XLarge rp;
		_XL_FracUpper(&rp, r, blen + 1);
		xl_sub(t, &rp);
		DEBUG_Print("-R^: ", t, "\n");
		if (xl_cmpzero(t) > 0)
		{
			--quot;
			xl_sub(t, nb);
			if (xl_cmpzero(t) > 0)
			{
				--quot;
				xl_sub(t, nb);
			}
		}
		xl_neg(t);
		DEBUG_Print("R: ", t, "\n");
		q->frac[q->begin + i - (blen - 1 + expDif)] = quot;

		_XL_GetDataW(r->frac + i + 2 - (blen + 1), t, 0, blen + 1);
		if (*_XL_GetFracBack(r) == 0)
		{
			_XL_PopBack(r);
		}
	}

	if(bs < 31) xl_shr(r, 31 - bs);
	_XL_Shrink(q);
	q->sign = a->sign ^ b->sign;
	xl_delete(t);
	xl_delete(nb);
}

XLarge _XL_Constructor(XL_UINT32 size, XL_UINT32 begin, XL_UINT32 end, XL_UINT32* frac, XL_UINT32 sign, XL_INT32 exp)
{
	XLarge x;
	x.size = size;
	x.begin = begin;
	x.end = end;
	x.frac = frac;
	x.sign = sign;
	x.exp = exp;
	return x;
}

void _i_mod32(XL_UINT32* aFrac, XL_UINT32 aLen, const XL_UINT32* bFrac, XL_UINT32 bLen, XL_UINT32* q, PXLARGE t);


void _i_mod42(XL_UINT32* aFrac, XL_UINT32 aLen, const XL_UINT32* bFrac, XL_UINT32 bLen, XL_UINT32* q, PXLARGE t)
{
	if (bLen == 1)
	{
		if (aLen < 2)
		{
			*q = *aFrac / *bFrac;
			*aFrac = *aFrac % *bFrac;
		}
		else
		{
			XL_UINT32 rem = aFrac[aLen - 1];
			q[aLen - 1] += rem / *bFrac;
			rem = rem % *bFrac;
			for (XL_UINT32 i = aLen - 2; i != (XL_UINT32)(-1); --i)
			{
				XL_UINT32 d;
				rem = i_div64(i_64(rem, aFrac[i]), *bFrac, &q[i]);
			}
			memset(aFrac + 1, 0, sizeof(XL_UINT32)*(aLen - 1));
			*aFrac = rem;
		}
		return;
	}
	
	XL_UINT32 topQ = -1;
	if (aLen >= bLen)
	{
		if (i_cmps(&aFrac[aLen - bLen], bFrac, bLen) >= 0)
		{
			i_subs(&aFrac[aLen - bLen], bFrac, bLen, 0);
			topQ = aLen - bLen;
		}
		while (aLen > 0 && aFrac[aLen - 1] == 0) --aLen;
	}

	XL_UINT32 unitLen = (bLen + 1) / 2;

	if (aLen > bLen + 1)
	{
		if ((aLen - bLen) % unitLen)
		{
			aLen += unitLen - ((aLen - bLen) % unitLen);
		}
		while (aLen > bLen)
		{
			XL_UINT32 tLen = MIN(aLen, unitLen + bLen);
			_i_mod32(aFrac + aLen - tLen, tLen, bFrac, bLen, q + aLen - tLen, t);
			aLen -= tLen - bLen;
			while (aLen > 0 && aFrac[aLen - 1] == 0) --aLen;
		}
	}

	if (topQ != (XL_UINT32)-1) q[topQ] = 1;

	if (aLen >= bLen)
	{
		if (i_cmps(&aFrac[aLen - bLen], bFrac, bLen) >= 0)
		{
			i_subs(&aFrac[aLen - bLen], bFrac, bLen, 0);
			q[aLen - bLen] = 1;
		}
		while (aLen > 0 && aFrac[aLen - 1] == 0) --aLen;
		XLarge ax = _XL_Constructor(aLen, 0, aLen, aFrac, 0, 0);
		XLarge bx = _XL_Constructor(bLen, 0, bLen, (XL_UINT32*)bFrac, 0, 0);
		for (XL_UINT32 i = aLen - 2; i >= bLen - 1; --i)
		{
			XL_UINT32 quot;
			i_div64(*(XL_UINT64*)&aFrac[i], bFrac[bLen - 1], &quot);
			xl_mul1(t, &bx, quot);
			xl_sub(&ax, t);
			if (xl_cmpzero(&ax) < 0)
			{
				--quot;
				xl_add(&ax, &bx);
				if (xl_cmpzero(&ax) < 0)
				{
					--quot;
					xl_add(&ax, &bx);
				}
			}
			q[i - (bLen-1)] = quot;
		}
	}
}

void _i_mod32(XL_UINT32* aFrac, XL_UINT32 aLen, const XL_UINT32* bFrac, XL_UINT32 bLen, XL_UINT32* q, PXLARGE t)
{
	XL_UINT32 unitLen = (bLen + 1) / 2;
	XL_UINT32 tLen = MIN(aLen, unitLen * 2);
	
	if (tLen == unitLen*2 && i_cmps(&aFrac[aLen - unitLen], &bFrac[bLen - unitLen], unitLen) >= 0)
	{
		memset(q, 0xFFFFFFFF, sizeof(XL_UINT32)*unitLen);
		i_subs(&aFrac[aLen - unitLen], &bFrac[bLen - unitLen], unitLen, 0);
		if (i_adds(&aFrac[aLen - tLen], &bFrac[bLen - unitLen], unitLen, 0))
		{
			i_addu32(&aFrac[aLen - unitLen], unitLen, 1);
		}
	}
	else
	{
		_i_mod42(aFrac + aLen - tLen, tLen, bFrac + bLen - unitLen, unitLen, q, t);
	}
	
	XLarge ax = _XL_Constructor(aLen + 2, 0, aLen, aFrac, 0, 0);
	XLarge bx = _XL_Constructor(bLen, 0, bLen, (XL_UINT32*)bFrac, 0, 0);
	XLarge b0x = _XL_Constructor(bLen, 0, bLen - unitLen, (XL_UINT32*)bFrac, 0, 0);
	XLarge qx = _XL_Constructor(aLen - bLen, 0, aLen - bLen, q, 0, 0);

	xl_mul(t, &b0x, &qx);
	xl_sub(&ax, t);
	if (xl_cmpzero(&ax) < 0)
	{
		xl_sub1(&qx, 1);
		xl_add(&ax, &bx);
		if (xl_cmpzero(&ax) < 0)
		{
			xl_sub1(&qx, 1);
			xl_add(&ax, &bx);
		}
	}
}

void _XL_ModNM(PCXLARGE a, PCXLARGE b, PXLARGE q, PXLARGE r)
{
	XL_UINT32 bs = i_bsu(*_XL_GetFracBack(b));
	if (bs == (XL_UINT32)(-1))
	{
		// b 가 0인 경우
		xl_setnan(q);
		xl_setnan(r);
		return;
	}

	switch (_Abs_Cmp(a, b))
	{
	case 0:
		xl_init1(q, 1);
		xl_init1(r, 0);
		return;
	case -1:
		xl_init1(q, 0);
		xl_move(r, a);
		return;
	}

	_XL_ExpandCopy(r, a, MAX(a->exp - b->exp, 0), _XL_GetLength(b)+1);
	PXLARGE nb = xl_createcopy(b);

	if (bs < 31)
	{
		_XL_SHL(r, 31 - bs);
		xl_shl(nb, 31 - bs);
	}

	while (*_XL_GetFracBack(r) == 0)
	{
		_XL_PopBack(r);
	}

	XL_INT32 expDif = nb->exp - r->exp;

	XL_UINT32 alen = _XL_GetLength(r);
	XL_UINT32 blen = _XL_GetLength(nb);
	XLarge bAbs = *b;
	bAbs.sign = 0;
	r->sign = 0;
	xl_init1(q, 0);
	_XL_PushBackN(q, alen);
	PXLARGE t = xl_create(alen + blen);

	_i_mod42(r->frac + r->begin + expDif, alen - expDif, nb->frac + nb->begin, blen, q->frac + q->begin, t);

	if (bs < 31) xl_shr(r, 31 - bs);
	_XL_Shrink(q);
	q->sign = a->sign ^ b->sign;
	xl_delete(t);
	xl_delete(nb);

	//DEBUG_Print("", a, " / ");
	//DEBUG_Print("", b, " = ");
	//DEBUG_Print("\n", q, " .. ");
	//DEBUG_Print("", r, "\n");
}

void xl_mod(PCXLARGE dest, PCXLARGE divisor, PXLARGE quot, PXLARGE remainder)
{
	return _XL_ModNM(dest, divisor, quot, remainder);
}

/*void xl_mod(PXLARGE dest, PXLARGE divisor, PXLARGE quot)
{
	if(xl_isnan(dest) || xl_isnan(divisor)) return xl_setnan(quot);
	XL_UINT32 bsDivisor = i_bsu(*_XL_GetFracBack(divisor));
	// 나누는 수가 0일 경우
	if(bsDivisor == (XL_UINT32)-1)
	{
		xl_setnan(quot);
		return;
	}

	XL_UINT32 destLen = _XL_GetLength(dest);
	XL_UINT32 divLen = _XL_GetLength(divisor);
	XL_UINT32 expBias = divisor->exp - dest->exp;
	if(dest->exp > divisor->exp)
	{
		expBias = 0;
		// 아랫쪽으로 확장
		_XL_PushFrontN(dest, dest->exp - divisor->exp);
		destLen = _XL_GetLength(dest);
	}

	// 나누는 수가 더 크면 몫은 0
	if(dest->exp + destLen < divisor->exp + divLen)
	{
		xl_init(quot, 0);
		return;
	}
	XL_INT32 bit = ((dest->exp + destLen) - (divisor->exp + divLen)) << 5;
	bit += i_bsu(*_XL_GetFracBack(dest)) - bsDivisor;
	if(bit < 0)
	{
		xl_init(quot, 0);
		return;
	}

	xl_shl(divisor, bit);
	xl_init(quot, 0);
	_XL_PushBackN(quot, bit >> 5);
	
	for(XL_UINT32 i = bit;;--i)
	{
		if(_Abs_Cmp(dest, divisor) >= 0)
		{
			_Abs_Sub(dest, divisor);
			*_XL_GetFrac(quot, i >> 5) |= 1 << (i & 31);
		}
		if(i == 0) break;
		xl_shr(divisor, 1);
	}
	quot->sign = dest->sign ^ divisor->sign;
	_XL_Shrink(quot);
}*/

XL_UINT32 xl_mod1(PCXLARGE dest, XL_UINT32 divisor, PXLARGE quot)
{
	if (xl_isnan(dest) || divisor == 0) return xl_setnan(quot), 0;
	XL_UINT32 len = _XL_GetLength(dest);
	if (len < 2)
	{
		XL_UINT32 a = *_XL_GetFrac(dest, 0);
		xl_init1(quot, a / divisor);
		return a % divisor;
	}
	else
	{
		xl_init1(quot, 0);
		_XL_PushBackN(quot, len - 1);
		XL_UINT32 rem = *_XL_GetFrac(dest, len - 1);
		quot->frac[len - 1] = rem / divisor;
		rem = rem % divisor;
		for (XL_UINT32 i = len - 2; i != (XL_UINT32)(-1); --i)
		{
			XL_UINT32 d;
			rem = i_div64(i_64(rem, *_XL_GetFrac(dest, i)), divisor, &quot->frac[i]);
		}
		_XL_Shrink(quot);
		return rem;
	}
}

inline void _XL_PrintInt(PXLARGE pxl, PXLARGE unit, FILE* file)
{
	PXLARGE quot = xl_create(4);
	PXLARGE r = xl_create(4);
	xl_init(quot, 0);
	if((int)_XL_GetLength(pxl) + pxl->exp <= 0)
	{
		fputc('0', file);
		return;
	}

	XL_UINT32 decimalSize = (XL_UINT32)((_XL_GetLength(pxl) + pxl->exp) * 1.07033) + 1;
	XL_UINT32* decimals = (XL_UINT32*)malloc(sizeof(XL_UINT32) * decimalSize);

	for(XL_UINT32 i = 0; i < decimalSize; ++i)
	{
		xl_mod(pxl, unit, quot, r);
		decimals[i] = *_XL_GetFrac(r, 0);
		xl_swap(pxl, quot);
	}
	
	XL_UINT32 i = decimalSize - 1;
	for(; i != (XL_UINT32)-1; --i)
	{
		if(decimals[i] || i == 0) break;
	}
	fprintf(file, "%d", decimals[i--]);
	for(; i != (XL_UINT32)-1; --i)
	{
		fprintf(file, "%09d", decimals[i]);
	}

	free(decimals);
	xl_delete(r);
	xl_delete(quot);
}

inline void _XL_PrintFrac(PXLARGE pxl, PCXLARGE unit, FILE* file, XL_UINT32 precision)
{
	PXLARGE result = xl_create(4);

	XL_UINT32 decimalSize = (XL_UINT32)(
		(precision == (XL_UINT32)-1 ? (-pxl->exp) : (precision)
		) * 1.07033);

	for(XL_UINT32 i = 0; i < decimalSize; ++i)
	{
		xl_mul(result, pxl, unit);
		if((XL_INT32)(_XL_GetLength(result) + result->exp) > 0)
		{
			fprintf(file, "%09d", *_XL_GetFracBack(result));
			if(_XL_PopBack(result)) break;
		}
		else
		{
			fprintf(file, "000000000");
		}
		xl_swap(result, pxl);
	}

	xl_delete(result);
}

void xl_print(PCXLARGE pxl, FILE* file, XL_UINT32 precision)
{
	if(xl_isnan(pxl)) return (void)fprintf(file, "NaN");
	PXLARGE integer = xl_create(pxl->size/2 + 1);
	PXLARGE fraction = xl_create(pxl->size/2 + 1);

	xl_modf(integer, fraction, pxl);

	PXLARGE unit = xl_create(2);
	xl_init(unit, 1000000000);

	if(pxl->sign) fputc('-', file);
	_XL_PrintInt(integer, unit, file);
	if(*_XL_GetFracBack(fraction))
	{
		XL_UINT32 fprec = precision;
		if(fprec != (XL_UINT32)-1)
		{
			XL_UINT32 intLen = _XL_GetLength(integer);
			if(fprec > intLen)
			{
				fprec -= intLen;
			}
			else
			{
				fprec = 1;
			}
		}
		fputc('.', file);
		_XL_PrintFrac(fraction, unit, file, fprec);
	}

	xl_delete(unit);
	xl_delete(fraction);
	xl_delete(integer);
}

int xl_isinteger(PCXLARGE dest)
{
	return dest->exp >= 0;
}

void xl_modf(PXLARGE integer, PXLARGE fraction, PCXLARGE pxl)
{
	if(xl_isnan(pxl)) return xl_setnan(integer), xl_setnan(fraction);
	XL_UINT32 len = _XL_GetLength(pxl);
	if(fraction)
	{
		xl_init(fraction, 0);
		if(pxl->exp < 0)
		{
			XL_UINT32 fracSize = -pxl->exp;
			if(fracSize > len) fracSize = len;
			_XL_PushBackN(fraction, fracSize - 1);
			if(pxl->begin + fracSize <= pxl->size)
			{
				memcpy(&fraction->frac[fraction->begin], 
					&pxl->frac[pxl->begin], fracSize * sizeof(XL_UINT32));
			}
			else
			{
				XL_UINT32 sep = pxl->size - pxl->begin;
				memcpy(&fraction->frac[fraction->begin], 
					&pxl->frac[pxl->begin], sep * sizeof(XL_UINT32));
				memcpy(&fraction->frac[fraction->begin + sep], 
					pxl->frac, (fracSize - sep) * sizeof(XL_UINT32));
			}
			fraction->exp = pxl->exp;
			_XL_Shrink(fraction);
		}
	}
	if(integer)
	{
		xl_init(integer, 0);
		if((XL_INT32)(pxl->exp + len) > 0)
		{
			XL_UINT32 intSize = pxl->exp + len;
			if(intSize > len) intSize = len;
			XL_UINT32 point = 0;
			if(pxl->exp < 0) point = -pxl->exp;
			point += pxl->begin;
			if(point > pxl->size) point -= pxl->size;
			_XL_PushBackN(integer, intSize - 1);
			if(point + intSize <= pxl->size)
			{
				memcpy(&integer->frac[integer->begin], 
					&pxl->frac[point], intSize * sizeof(XL_UINT32));
			}
			else
			{
				XL_UINT32 sep = pxl->size - point;
				memcpy(&integer->frac[integer->begin], 
					&pxl->frac[point], sep * sizeof(XL_UINT32));
				memcpy(&integer->frac[integer->begin + sep], 
					pxl->frac, (intSize - sep) * sizeof(XL_UINT32));
			}
			integer->exp = pxl->exp;
			if(integer->exp < 0) integer->exp = 0;
			_XL_Shrink(integer);
		}
	}
}

void xl_round(PXLARGE dest, XL_UINT32 precision)
{
	XL_UINT32 len = _XL_GetLength(dest);
	if(len <= precision) return;
	if(*_XL_GetFrac(dest, len - precision - 1) > 0x80000000)
	{
		_XL_PushBack(dest);
		XL_UINT32 pos = dest->begin + len - precision;
		if(dest->begin < dest->end)
		{
			i_addu32(&dest->frac[pos], dest->end - pos, 1);
		}
		else
		{
			if(pos < dest->size)
			{
				if(i_addu32(&dest->frac[pos], dest->size - (pos), 1))
				{
					if(dest->end) i_addu32(dest->frac, dest->end, 1);
				}
			}
			else
			{
				i_addu32(&dest->frac[pos - dest->size], dest->end - (pos - dest->size), 1);
			}
		}
	}
	_XL_PopFrontN(dest, len - precision);
	_XL_Shrink(dest);
}

void xl_inv(PXLARGE dest, PCXLARGE a, XL_UINT32 precision)
{
	if(xl_isnan(a)) return xl_setnan(dest);
	XL_UINT32 aLen = _XL_GetLength(a);
	XL_UINT32 bsA = i_bsu(*_XL_GetFracBack(a));
	// a가 0일 경우
	if(bsA == (XL_UINT32)-1)
	{
		xl_setnan(dest);
		return;
	}

	XL_UINT32 bits = ((aLen - 1) << 5) + bsA;
	XL_INT32 exp = a->exp * 32 + bits;

	PXLARGE na = xl_createcopy(a);
	na->sign = 0;
	PXLARGE t = xl_create(a->size);
	PXLARGE u = xl_create(a->size);
	//PXLARGE ttt = xl_create(a->size);

	if(exp < -1)
	{
		xl_shl(na, -(exp + 1));
	}
	else if(exp >= 0)
	{
		xl_shr(na, exp + 1);
	}

	xl_init(dest, 3);
	xl_init(u, 2);
	xl_mulsub(dest, na, u);
	
	//XL_UINT32 prec = _XL_GetLength(dest);
	while(1)
	{
		xl_init(t, 1);
		xl_mulsub(t, na, dest);
		if(xl_cmpnearzero(t, -(XL_INT32)precision) == 0) break;
		xl_round(t, precision + 2);
		xl_mul(u, t, dest);
		xl_round(u, precision + 2);
		xl_add(dest, u);
	}
	xl_round(dest, precision);

	if(exp < -1)
	{
		xl_shl(dest, -(exp + 1));
	}
	else if(exp >= 0)
	{
		xl_shr(dest, exp + 1);
	}

	dest->sign = a->sign;

	xl_delete(u);
	xl_delete(t);
	xl_delete(na);
}

XL_UINT32 powi(XL_UINT32 a, XL_UINT32 b)
{
	if(b == 0) return 1;
	if(b == 1) return a;
	int r = powi(a, b/2);
	if(b & 1) return r*r*a;
	else return r*r;
}

void xl_atof(PXLARGE dest, const char* str, XL_UINT32 precision)
{
	xl_init(dest, 0);
	PXLARGE t = xl_create(2);
	PXLARGE unit = xl_create(2);
	xl_init(unit, 1000000000);

	XL_UINT32 sign = 0;
	//XL_UINT32 buf = 0;
	
	char buf[10] = {0};
	int bufSize = 0;

	switch(*str)
	{
	case '-':
		sign = 1;
		++str;
		break;
	case '+':
		++str;
		break;
	}

	for(;iswdigit(*str);++str)
	{
		buf[bufSize++] = *str;
		if(bufSize == 9)
		{
			buf[bufSize] = 0;
			bufSize = 0;
			xl_init(t, atoi(buf));
			xl_muladd(t, dest, unit);
			xl_swap(t, dest);
		}
	}
	if(bufSize)
	{
		buf[bufSize] = 0;
		xl_init(t, atoi(buf));
		xl_init(unit, powi(10, bufSize));
		bufSize = 0;
		xl_muladd(t, dest, unit);
		xl_swap(t, dest);
	}

#define NSIZE 9
#define PINC 1

	if(*str == '.')
	{
		size_t strLen = 0;
		const char* dot = str;
		++str;
		for(;iswdigit(*str);++str)
		{
			strLen++;
		}
		--str;
		if(precision == (XL_UINT32)(-1))
		{
			precision = (strLen + 8) / 9;
		}

		XL_UINT32 p = 2 + precision - (strLen / 9);
		PXLARGE decimal = xl_create(2);
		xl_init(decimal, 0);
		xl_init(t, powi(10, NSIZE));
		xl_inv(unit, t, p);
		p += PINC;

		
		buf[NSIZE] = 0;
		for(;str != dot;--str)
		{
			buf[NSIZE - ++bufSize] = *str;
			if(bufSize == NSIZE)
			{
				xl_init(t, powi(10, NSIZE));
				xl_inv(unit, t, p);
				p += PINC;
				bufSize = 0;
				xl_init(t, atoi(buf));
				xl_add(decimal, t);
				xl_mul(t, decimal, unit);
				xl_swap(t, decimal);
			}
		}
		if(bufSize)
		{
			xl_init(t, powi(10, bufSize));
			xl_inv(unit, t, p + ((bufSize*2 > NSIZE) ? PINC : 0));
			xl_init(t, atoi(buf + (NSIZE - bufSize)));
			xl_add(decimal, t);
			xl_mul(t, decimal, unit);
			xl_swap(t, decimal);
		}
		xl_round(decimal, precision + 1);
		xl_add(dest, decimal);
		xl_delete(decimal);
	}
	dest->sign = sign;
	xl_delete(t);
	xl_delete(unit);
}

/*XL_UINT64 _XL_Hash(PXLARGE pxl, XL_UINT32 targetPrecision)
{
	XL_UINT64 hash = 0;
	for(XL_UINT32 i = 0; i < targetPrecision; ++i)
	{
		hash += ((XL_UINT64)*_XL_GetFrac(pxl, i)) << (i & 31);
	}
	return hash;
}

void _XL_Sqrt(PXLARGE dest, PCXLARGE a, XL_UINT32 precision)
{
	if(xl_isnan(a)) return xl_setnan(dest);
	if(a->sign)
	{
		xl_setnan(dest);
		return;
	}
	XL_INT32 exp = (XL_INT32)_XL_GetLength(a) + a->exp - 1;
	XL_UINT32 estBit = i_bsu(*_XL_GetFracBack(a)) / 2 + (exp % 2 ? 16 : 0);
	xl_init1(dest, 1 << estBit);
	dest->exp = exp / 2 - (exp < 0 ? 1 : 0);

	PXLARGE t = xl_create(dest->size);

	const XL_UINT32 maxPrec = MAX(precision * 3 /2, precision + 8);
	XL_UINT32 prec = 2;
	XL_UINT32 testCnt = 0;
	XL_UINT64 errHash[2] = {0, };
	while(1)
	{
		//DEBUG_Print("x: ", dest, "\n");
		xl_inv(t, dest, prec + 1); // t = 1 / x;
		//DEBUG_Print("1/x: ", t, "\n");
		xl_muladd(dest, t, a);
		xl_shr(dest, 1); // nx = (x + a/x) / 2;
		xl_round(dest, prec);
		prec = MIN(prec * 3 / 2, maxPrec);
		//DEBUG_Print("nx: ", dest, "\n");
		if(prec >= maxPrec)
		{
			XL_UINT64 nHash = _XL_Hash(dest, MIN(precision, _XL_GetLength(dest)));
			if(testCnt > 0 && errHash[0] == nHash) break;
			if(testCnt > 1 && errHash[1] == nHash) break;
			errHash[1] = errHash[0];
			errHash[0] = nHash;
			++testCnt;
		}
	}
	xl_round(dest, precision);

	xl_delete(t);
}*/

inline void _XL_Sqrt_2(PXLARGE dest, PCXLARGE a, XL_UINT32 precision)
{
	if(xl_isnan(a)) return xl_setnan(dest);
	if(xl_cmpzero(a) < 0)
	{
		return xl_setnan(dest);
	}
	else if(xl_cmpzero(a) == 0)
	{
		return xl_init(dest, 0);
	}

	XL_INT32 exp = (XL_INT32)_XL_GetLength(a) + a->exp - 1;
	XL_UINT32 estBit = i_bsu(*_XL_GetFracBack(a)) / 2 + (exp % 2 ? 16 : 0);
	xl_init1(dest, 1 << (31 - estBit));
	dest->exp = exp / 2 - (exp < 0 ? 1 : 0);
	dest->exp = -dest->exp - 1;
	
	PXLARGE t = xl_create(precision);
	PXLARGE u = xl_create(precision);
	PXLARGE c2 = xl_create(2);
	xl_init(c2, 2);

	const XL_UINT32 maxPrec = MAX(precision * 4 / 3, precision + 8);
	XL_UINT32 prec = 2;
	//XL_UINT32 testCnt = 0;
	//XL_UINT64 errHash[2] = {0, };
	while(1)
	{
		xl_mul(t, dest, dest);
		xl_round(t, prec);
		xl_mul(u, t, a);
		xl_round(u, prec);
		xl_init(t, 1);
		xl_sub(t, u); // t = 1 - ax^2
		if(xl_cmpnearzero(t, -(XL_INT32)precision) == 0) break;
		xl_add(t, c2);
		xl_mul(u, t, dest); // u = x(3-ax^2)
		xl_shr(u, 1); // u = x(3-ax^2)/2
		xl_swap(u, dest);
		xl_round(dest, prec);
		prec = MIN(prec * 3 / 2, maxPrec);
		/*if(prec >= maxPrec)
		{
			XL_UINT64 nHash = _XL_Hash(dest, MIN(precision, _XL_GetLength(dest)));
			if(testCnt > 0 && errHash[0] == nHash) break;
			if(testCnt > 1 && errHash[1] == nHash) break;
			errHash[1] = errHash[0];
			errHash[0] = nHash;
			++testCnt;
			//if(++testCnt >= 8) break;
		}*/
	}
	//printf("testCnt:%d\n", testCnt);
	xl_mul(u, dest, a);
	xl_swap(u, dest);
	xl_round(dest, precision);

	xl_delete(c2);
	xl_delete(u);
	xl_delete(t);
}

void xl_sqrt(PXLARGE dest, PCXLARGE a, XL_UINT32 precision)
{
	return _XL_Sqrt_2(dest, a, precision);
}

inline void _XL_Exp(PXLARGE dest, PCXLARGE a, XL_UINT32 precision)
{
	PXLARGE t = xl_create(precision*2);
	PXLARGE u = xl_create(precision*2);
	PXLARGE rcp = xl_create(precision*2);

	xl_init(dest, 1);
	xl_add(dest, a);
	xl_move(t, a);

	for(XL_UINT32 i = 2; ; ++i)
	{
		xl_init(u, i);
		xl_inv(rcp, u, precision + 1);
		xl_mul(u, t, a);
		xl_mul(t, u, rcp); // t *= a / i;
		xl_round(t, precision + 2);
		XL_INT32 destHigh = dest->exp + (XL_INT32)_XL_GetLength(dest) - 1;
		if(xl_cmpnearzero(t, destHigh - (XL_INT32)precision) == 0) break;
		xl_add(dest, t);
	}
	xl_round(dest, precision);

	xl_delete(rcp);
	xl_delete(u);
	xl_delete(t);
}

void xl_exp(PXLARGE dest, PCXLARGE a, XL_UINT32 precision)
{
	if(xl_isnan(a)) return xl_setnan(dest);
	XL_UINT32 aLen = _XL_GetLength(a);
	XL_UINT32 bsA = i_bsu(*_XL_GetFracBack(a));
	// a가 0일 경우
	if(bsA == (XL_UINT32)-1)
	{
		xl_init(dest, 1);
		return;
	}

	XL_UINT32 bits = ((aLen - 1) << 5) + bsA;
	XL_INT32 exp = a->exp * 32 + bits;
	exp += 32;
	// a의 절대값이 이미 충분히 작으면 단순 Exp계산
	if(exp < 0)
	{
		return _XL_Exp(dest, a, precision);
	}

	// a를 1보다 작게 시프트
	PXLARGE na = xl_createcopy(a);
	xl_shr(na, exp + 1);

	CPUTimer timer;
	_XL_Exp(dest, na, precision + exp + 1);
	timer.Print(stdout, "xl_exp._Exp:\t%.4fs\n");

	timer.Reset();
	for(XL_INT32 i = 0; i <= exp; ++i)
	{
		xl_mul(na, dest, dest);
		xl_swap(na, dest);
		xl_round(dest, precision + exp - i);
	}
	timer.Print(stdout, "xl_exp.Sqrs:\t%.4fs\n");

	xl_delete(na);
}

inline void _XL_Log(PXLARGE dest, PCXLARGE a, XL_UINT32 precision)
{
	PXLARGE za = xl_createcopy(a);
	PXLARGE zb = xl_createcopy(a);
	PXLARGE t = xl_create(precision * 2);
	PXLARGE rcp = xl_create(precision * 2);
	
	xl_init(t, 1);
	xl_add(za, t);
	xl_sub(zb, t);
	xl_inv(t, za, precision + 2);
	xl_mul(za, zb, t); // za = (a-1)/(a+1)
	xl_round(za, precision + 2);
	xl_mul(zb, za, za); // zb = {(a-1)/(a+1)}^2
	xl_round(zb, precision + 2);
	xl_move(t, za); // t = (a-1)/(a+1)
	xl_move(dest, t);

	//DEBUG_Print("zb: ", zb, "\n");

	for(XL_UINT32 i = 3; ; i += 2)
	{
		//DEBUG_Print("\ndest: ", dest, "\n");
		xl_init(za, i);
		xl_inv(rcp, za, precision + 1);
		//DEBUG_Print("rcp: ", rcp, "\n");
		xl_mul(za, t, zb);
		xl_swap(za, t); // t *= {(a-1)/(a+1)}^2
		xl_round(t, precision + 2);
		//DEBUG_Print("t: ", t, "\n");
		xl_mul(za, t, rcp);
		xl_round(za, precision + 2);
		//DEBUG_Print("za: ", za, "\n");
		XL_INT32 destHigh = dest->exp + (XL_INT32)_XL_GetLength(dest) - 1;
		if(xl_cmpnearzero(za, destHigh - (XL_INT32)precision) == 0) break;
		xl_add(dest, za);
	}
	xl_shl(dest, 1);
	xl_round(dest, precision);

	xl_delete(rcp);
	xl_delete(t);
	xl_delete(za);
	xl_delete(zb);
}

void xl_log(PXLARGE dest, PCXLARGE a, XL_UINT32 precision)
{
	if(xl_isnan(a)) return xl_setnan(dest);
	// a가 0 혹은 음수일 경우
	if(xl_cmpzero(a) <= 0)
	{
		xl_setnan(dest);
		return;
	}

	XL_UINT32 aLen = _XL_GetLength(a);
	XL_UINT32 bsA = i_bsu(*_XL_GetFracBack(a));
	XL_UINT32 bits = ((aLen - 1) << 5) + bsA;
	XL_INT32 exp = a->exp * 32 + bits;

	XL_UINT32 qsqrt = i_bsu(abs(exp)) + 1;
	qsqrt += 16;
	if(qsqrt == 0)
	{
		return _XL_Log(dest, a, precision);
	}
	PXLARGE na = xl_createcopy(a);
	PXLARGE t = xl_create(precision + qsqrt);

	CPUTimer timer;
	for(XL_UINT32 i = 0; i < qsqrt; ++i)
	{
		//DEBUG_Print("na: ", na, "\n\n");
		xl_sqrt(t, na, precision + (i + 1));
		xl_swap(t, na);
	}
	timer.Print(stdout, "xl_log.Sqrts:\t%.4fs\n");

	timer.Reset();
	_XL_Log(dest, na, precision + 1);
	timer.Print(stdout, "xl_log._Log:\t%.4fs\n");

	xl_shl(dest, qsqrt);
	xl_round(dest, precision);

	xl_delete(t);
	xl_delete(na);
}

inline void _XL_Powi(PXLARGE dest, PCXLARGE base, PCXLARGE exp, XL_UINT32 precision)
{
	if(xl_cmpzero(base) == 0)
	{
		if(xl_cmpzero(exp) > 0)
		{
			return xl_init(dest, 0);
		}
		else
		{
			return xl_setnan(dest);
		}
	}

	if(xl_cmpzero(exp) == 0)
	{
		return xl_init(dest, 1);
	}

	PXLARGE nbase = (PXLARGE)base;
	if(xl_cmpzero(exp) < 0)
	{
		nbase = xl_create(precision);
		xl_inv(nbase, base, precision);
	}

	PXLARGE t = xl_create(precision * 2);
	xl_init(dest, 1);
	xl_init(t, 1);
	XL_UINT32 len = _XL_GetLength(exp);
	XL_UINT32 lenBit = len << 5;
	for(XL_UINT32 i = 0; i < lenBit; ++i)
	{
		xl_mul(t, dest, dest);
		xl_round(t, precision + 1);
		XL_UINT32 d = *_XL_GetFrac(exp, len - 1 - (i >> 5));
		if((d >> (31 - (i & 31))) & 1)
		{
			xl_mul(dest, t, nbase);
			xl_round(dest, precision + 1);
		}
		else
		{
			xl_swap(t, dest);
		}
	}
	xl_delete(t);
	if(xl_cmpzero(exp) < 0)
	{
		xl_delete(nbase);
	}
	xl_round(dest, precision);
}

void xl_pow(PXLARGE dest, PCXLARGE base, PCXLARGE exp, XL_UINT32 precision)
{
	if(xl_isinteger(exp))
	{
		return _XL_Powi(dest, base, exp, precision);
	}

	// base가 음수인 경우
	if(xl_cmpzero(base) < 0)
	{
		return xl_setnan(dest);
	}

	// base가 0인 경우
	if(xl_cmpzero(base) == 0)
	{
		if(xl_cmpzero(exp) > 0)
		{
			return xl_init(dest, 0);
		}
		else
		{
			return xl_setnan(dest);
		}
	}


	PXLARGE l = xl_create(precision);
	xl_log(l, base, precision);
	xl_mul(dest, l, exp);
	xl_swap(dest, l);
	xl_exp(dest, l, precision);
	xl_delete(l);
}

inline void _XL_Cos(PXLARGE dest, PCXLARGE rad, XL_UINT32 precision)
{
	PXLARGE r2 = xl_create(precision*2);
	PXLARGE t = xl_create(precision*2);
	PXLARGE u = xl_create(precision*2);
	PXLARGE rcp = xl_create(precision*2);

	xl_init(dest, 1);
	xl_init(t, 1);
	xl_mul(r2, rad, rad);
	xl_round(r2, precision + 1);

	for(XL_UINT32 i = 1; ; ++i)
	{
		xl_init1(u, (2*i - 1) * (2*i));
		u->sign = 1;
		xl_inv(rcp, u, precision + 1);
		//DEBUG_Print("rcp: ", rcp, "\n");
		xl_mul(u, t, r2);
		xl_mul(t, u, rcp); // t *= rad^2 / ((2i-1)2i);
		xl_round(t, precision + 2);
		//DEBUG_Print("t: ", t, "\n");
		XL_INT32 destHigh = dest->exp + (XL_INT32)_XL_GetLength(dest) - 1;
		if(xl_cmpnearzero(t, destHigh - (XL_INT32)precision) == 0) break;
		xl_add(dest, t);
	}
	xl_round(dest, precision);

	xl_delete(rcp);
	xl_delete(u);
	xl_delete(t);
	xl_delete(r2);
}

void xl_cos(PXLARGE dest, PCXLARGE rad, XL_UINT32 precision)
{
	if(xl_isnan(rad)) return xl_setnan(dest);
	XL_UINT32 aLen = _XL_GetLength(rad);
	XL_UINT32 bsA = i_bsu(*_XL_GetFracBack(rad));
	// rad가 0일 경우
	if(bsA == (XL_UINT32)-1)
	{
		xl_init(dest, 1);
		return;
	}

	XL_UINT32 bits = ((aLen - 1) << 5) + bsA;
	XL_INT32 exp = rad->exp * 32 + bits;
	exp += 32;
	// rad의 절대값이 이미 충분히 작으면 단순 Cos계산
	if(exp < 0)
	{
		return _XL_Cos(dest, rad, precision);
	}

	PXLARGE one = xl_create(2);
	xl_init(one, 1);
	// rad를 1보다 작게 시프트
	PXLARGE na = xl_createcopy(rad);
	xl_shr(na, exp + 1);

	CPUTimer timer;
	_XL_Cos(dest, na, precision + exp + 1);
	timer.Print(stdout, "xl_cos._Cos:\t%.4fs\n");

	timer.Reset();
	for(XL_INT32 i = 0; i <= exp; ++i)
	{
		xl_mul(na, dest, dest);
		xl_shl(na, 1);
		xl_sub(na, one);
		xl_swap(na, dest);
		xl_round(dest, precision + exp - i);
	}
	timer.Print(stdout, "xl_cos.Sqrs:\t%.4fs\n");

	xl_delete(one);
	xl_delete(na);
}

inline void _XL_Sin(PXLARGE dest, PCXLARGE rad, XL_UINT32 precision)
{
	PXLARGE r2 = xl_create(precision*2);
	PXLARGE t = xl_create(precision*2);
	PXLARGE u = xl_create(precision*2);
	PXLARGE rcp = xl_create(precision*2);

	xl_move(dest, rad);
	xl_move(t, rad);
	xl_mul(r2, rad, rad);
	xl_round(r2, precision + 1);

	for(XL_UINT32 i = 1; ; ++i)
	{
		xl_init1(u, (2*i) * (2*i + 1));
		u->sign = 1;
		xl_inv(rcp, u, precision + 1);
		//DEBUG_Print("rcp: ", rcp, "\n");
		xl_mul(u, t, r2);
		xl_mul(t, u, rcp); // t *= rad^2 / ((2i+1)2i);
		xl_round(t, precision + 2);
		//DEBUG_Print("t: ", t, "\n");
		XL_INT32 destHigh = dest->exp + (XL_INT32)_XL_GetLength(dest) - 1;
		if(xl_cmpnearzero(t, destHigh - (XL_INT32)precision) == 0) break;
		xl_add(dest, t);
	}
	xl_round(dest, precision);

	xl_delete(rcp);
	xl_delete(u);
	xl_delete(t);
	xl_delete(r2);
}

void xl_sin(PXLARGE dest, PCXLARGE rad, XL_UINT32 precision)
{
	if(xl_isnan(rad)) return xl_setnan(dest);
	XL_UINT32 aLen = _XL_GetLength(rad);
	XL_UINT32 bsA = i_bsu(*_XL_GetFracBack(rad));
	// rad가 0일 경우
	if(bsA == (XL_UINT32)-1)
	{
		xl_init(dest, 0);
		return;
	}

	XL_UINT32 bits = ((aLen - 1) << 5) + bsA;
	XL_INT32 exp = rad->exp * 32 + bits;
	exp = exp * 0.63093 + 24;
	if(exp < 0)
	{
		return _XL_Sin(dest, rad, precision);
	}

	CPUTimer timer;
	PXLARGE t = xl_create(precision);
	PXLARGE u = xl_create(precision);
	PXLARGE v = xl_create(2);
	xl_init(u, 3);
	xl_inv(t, u, precision);
	xl_init(v, exp);
	_XL_Powi(u, t, v, precision);

	// rad 1보다 작게 만들기
	PXLARGE na = xl_create(precision);
	xl_mul(na, rad, u);
	//DEBUG_Print("na: ", na, "\n");
	timer.Print(stdout, "xl_sin.Normalize:\t%.4fs\n");

	timer.Reset();
	_XL_Sin(dest, na, precision + exp + 1);
	timer.Print(stdout, "xl_sin._Sin:\t%.4fs\n");

	timer.Reset();
	xl_init(v, 3);
	for(XL_INT32 i = 0; i < exp; ++i)
	{
		//DEBUG_Print("d: ", dest, "\n");
		xl_mul(na, dest, dest);
		xl_mul(u, na, dest);
		xl_shl(u, 2); // u = 4 sin^3 x
		xl_mul(t, dest, v); // t = 3 sin x
		xl_sub(t, u);
		xl_swap(t, dest);
		xl_round(dest, precision + exp - i);
	}
	xl_round(dest, precision);
	timer.Print(stdout, "xl_sin.Sqrs:\t%.4fs\n");

	xl_delete(na);
	xl_delete(v);
	xl_delete(u);
	xl_delete(t);
}



void xl_tan(PXLARGE dest, PCXLARGE rad, XL_UINT32 precision)
{
	if(xl_isnan(rad)) return xl_setnan(dest);
	PXLARGE sin, cos;
	sin = xl_create(precision + 1);
	cos = xl_create(precision + 1);

	xl_cos(sin, rad, precision + 1);
	xl_inv(cos, sin, precision + 1);
	xl_sin(sin, rad, precision);
	xl_mul(dest, sin, cos);
	xl_round(dest, precision);

	xl_delete(sin);
	xl_delete(cos);
}

inline void _XL_Atan(PXLARGE dest, PCXLARGE rad, XL_UINT32 precision)
{
	PXLARGE r2 = xl_create(precision*2);
	PXLARGE t = xl_create(precision*2);
	PXLARGE u = xl_create(precision*2);
	PXLARGE rcp = xl_create(precision*2);

	xl_move(dest, rad);
	xl_move(t, rad);
	xl_mul(r2, rad, rad);
	xl_round(r2, precision + 1);

	for(XL_UINT32 i = 1; ; ++i)
	{
		xl_init1(u, 2*i + 1);
		xl_inv(rcp, u, precision + 1);
		rcp->sign = i & 1;
		//DEBUG_Print("rcp: ", rcp, "\n");
		xl_mul(u, t, r2); 
		xl_round(u, precision + 2);
		xl_swap(u, t); // t *= rad^2
		xl_mul(u, t, rcp); // u = t / 2i
		xl_round(u, precision + 2);
		//DEBUG_Print("t: ", t, "\n");
		XL_INT32 destHigh = dest->exp + (XL_INT32)_XL_GetLength(dest) - 1;
		if(xl_cmpnearzero(u, destHigh - (XL_INT32)precision) == 0) break;
		xl_add(dest, u);
	}
	xl_round(dest, precision);

	xl_delete(rcp);
	xl_delete(u);
	xl_delete(t);
	xl_delete(r2);
}

void xl_atan(PXLARGE dest, PCXLARGE rad, XL_UINT32 precision)
{
	if(xl_isnan(rad)) return xl_setnan(dest);
	XL_UINT32 aLen = _XL_GetLength(rad);
	XL_UINT32 bsA = i_bsu(*_XL_GetFracBack(rad));
	// rad가 0일 경우
	if(bsA == (XL_UINT32)-1)
	{
		xl_init(dest, 0);
		return;
	}

	XL_UINT32 bits = ((aLen - 1) << 5) + bsA;
	XL_INT32 exp = rad->exp * 32 + bits;
	exp += 32;
	// rad의 절대값이 이미 충분히 작으면 단순 atan계산
	if(exp < 0)
	{
		return _XL_Atan(dest, rad, precision);
	}

	PXLARGE q = xl_create(precision * 2);
	PXLARGE one = xl_create(2);
	xl_init(one, 1);

	PXLARGE na = xl_create(precision * 2);
	xl_move(dest, rad);
	// rad를 1보다 작게 만들기
	CPUTimer timer;
	for(XL_INT32 i = 0; i <= exp; ++i)
	{
		xl_inv(q, dest, precision + i + 2);
		xl_mul(na, dest, dest);
		xl_add(na, one);
		xl_sqrt(dest, na, precision + i + 1);
		xl_sub(dest, one);
		xl_mul(na, dest, q);
		xl_swap(na, dest);
		xl_round(dest, precision + i + 1);
		//DEBUG_Print("r: ", dest, "\n");
	}
	timer.Print(stdout, "xl_atan.Halfs:\t%.4fs\n");
	
	timer.Reset();
	xl_move(na, dest);
	_XL_Atan(dest, na, precision + 1);
	xl_shl(dest, exp + 1);
	xl_round(dest, precision);
	timer.Print(stdout, "xl_atan._Atan:\t%.4fs\n");

	xl_delete(q);
	xl_delete(one);
	xl_delete(na);
}

void xl_asin(PXLARGE dest, PCXLARGE rad, XL_UINT32 precision)
{
	if(xl_isnan(rad)) return xl_setnan(dest);
	if(xl_cmpzero(rad) == 0)
	{
		return xl_init(dest, 0);
	}

	PXLARGE q = xl_create(precision * 2);
	PXLARGE one = xl_create(2);
	xl_init(one, 1);

	PXLARGE na = xl_create(precision * 2);
	xl_inv(q, rad, precision + 1);
	xl_init(na, 1);
	xl_mulsub(na, rad, rad);
	xl_sqrt(dest, na, precision + 1);
	xl_sub(dest, one);
	xl_neg(dest);
	xl_mul(na, dest, q); // na = (1-sqrt(1-rad^2)) / rad
	xl_round(na, precision + 1);
	//DEBUG_Print("r: ", dest, "\n");

	xl_atan(dest, na, precision + 1);
	xl_shl(dest, 1);
	xl_round(dest, precision);

	xl_delete(q);
	xl_delete(one);
	xl_delete(na);
}

void xl_acos(PXLARGE dest, PCXLARGE rad, XL_UINT32 precision)
{
	if(xl_isnan(rad)) return xl_setnan(dest);
	if(rad->sign == 1)
	{
		if(*_XL_GetFracBack(rad) == 1 && _XL_GetLength(rad) == 1)
		{
			return xl_pi(dest, precision);
		}
	}

	PXLARGE q = xl_create(precision * 2);
	PXLARGE one = xl_create(2);
	xl_init(one, 1);

	PXLARGE na = xl_create(precision * 2);
	xl_move(na, rad);
	xl_add(na, one);
	xl_inv(q, na, precision + 1);
	xl_mul(na, rad, rad);
	xl_sub(na, one);
	xl_neg(na);
	xl_sqrt(dest, na, precision + 1);
	xl_mul(na, dest, q); // na = sqrt(1 - rad^2) / (1 + rad)
	xl_round(na, precision + 1);
	//DEBUG_Print("r: ", dest, "\n");

	xl_atan(dest, na, precision + 1);
	xl_shl(dest, 1);
	xl_round(dest, precision);

	xl_delete(q);
	xl_delete(one);
	xl_delete(na);
}

void xl_pi(PXLARGE dest, XL_UINT32 precision)
{
	xl_setnan(dest);
	//
}